#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
#include <stdio.h>
#include <assert.h>

#include <pthread/rwlock_fcfs.h>

#ifdef DMALLOC
#include <dmalloc.h>
#endif

static const pthread_mutex_t initial_mutex_constant = PTHREAD_MUTEX_INITIALIZER;
static const pthread_cond_t initial_cond_constant = PTHREAD_COND_INITIALIZER;


pthread_rwlock_fcfs_t * pthread_rwlock_fcfs_alloc(void)
{
    pthread_rwlock_fcfs_t * rwlock;

    rwlock = malloc(sizeof(pthread_rwlock_fcfs_t));

    rwlock->queue = pthread_rwlock_fcfs_queue_alloc();

    /* We don't have any readers now. */
    rwlock->num_readers = 0;
    /* Much less a writer */
    rwlock->is_writer = 0;
    /* This RWLock is alive and kicking */
    rwlock->is_destroyed = 0;

    /* Initialize the mutex */
    rwlock->mutex = initial_mutex_constant;
    pthread_mutex_init(&(rwlock->mutex), NULL);

    return rwlock;
}

static int gain_write_generic(
    pthread_rwlock_fcfs_t * rwlock,
    int wait_for_access,
    int is_timed,
    const struct timespec * abstime,
    int (*continue_callback)(void * context),
    void * context
    )
{
    /* If there aren't any readers or writers in the queue we 
     * can gain access immidiately */
    if (rwlock->is_writer || (rwlock->num_readers > 0))
    {
        pthread_rwlock_fcfs_item_t * item;

        if (! wait_for_access)
        {
            return 1;
        }
        
        /* Someone is using the RWLock. Let's enqueue myself */
        item = malloc(sizeof(pthread_rwlock_fcfs_item_t));
        /* We are a writer, so let's designate ourselves as such */
        item->is_writer = 1;

        /* We want to gain access to the lock */
        item->is_disabled = 0;

        /* Initialize the condition variable */
        item->cond = initial_cond_constant;

        pthread_cond_init(&(item->cond), NULL);
    
        /* Let's put ourselves in the queue, so we will be release
         * when necessary. */
        pthread_rwlock_fcfs_queue_enqueue(rwlock->queue, item);
        
        /* Wait upon the condition variable to gain access to the rwlock */
        if (! is_timed)
        {
            pthread_cond_wait(&(item->cond), &(rwlock->mutex));
        }
        else
        {
            int ret;
           
            /* 
             * We try to get a timed wait on cond. If it times-out, we ask 
             * the user if he wishes to continue.
             */
            do
            {
                ret = pthread_cond_timedwait(&(item->cond), &(rwlock->mutex), abstime);
            } while 
                (
                    (ret == ETIMEDOUT) && 
                    (continue_callback != NULL) && 
                    continue_callback(context)
                );
           
            /* We failed for some reason (probably a time out - let's disable
             * this item, so it won't be processed.
             * */
            if (ret != 0)
            {
                item->is_disabled = 1;
                return ret;
            }
        }

        /* Now we have a write access */
        rwlock->is_writer = 1;

        return 0;
    }
    else
    {
        rwlock->is_writer = 1;

        return 0;
    }
}


void pthread_rwlock_fcfs_gain_write(pthread_rwlock_fcfs_t * rwlock PTHREAD_RWLOCK_FCFS_DEBUG_ARGS)
{
    pthread_mutex_lock(&(rwlock->mutex));

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
    printf("%s - Want Lock!\n", id);
#endif

    gain_write_generic(rwlock, 1, 0, NULL, NULL, NULL);

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
    printf("%s - Lock!\n", id);
#endif

    pthread_mutex_unlock(&(rwlock->mutex));
}

int pthread_rwlock_fcfs_try_gain_write(pthread_rwlock_fcfs_t * rwlock PTHREAD_RWLOCK_FCFS_DEBUG_ARGS)
{
    int ret;
    
    pthread_mutex_lock(&(rwlock->mutex));

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
    printf("%s - Want Try Lock!\n", id);
#endif

    ret = gain_write_generic(rwlock, 0, 0, NULL, NULL, NULL);

    if (ret == 0)
    {
#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
        printf("%s - Lock!\n", id);
#endif
    }
    else
    {
#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
        printf("%s - Failed Lock!\n", id);
#endif
    }

    pthread_mutex_unlock(&(rwlock->mutex));

    return ret;
}


/* This macro dequeues and frees a queue element */
#define dequeue_and_destroy_item() \
    pthread_cond_destroy(&(item->cond)); \
    free(item);     \
    pthread_rwlock_fcfs_queue_dequeue(rwlock->queue); \

#define extract_first_non_disabled_item() \
    /*       \
     * This loop extracts items from the queue until it encounters a  \
     * non-disabled item            \
     * */ \
    item = (pthread_rwlock_fcfs_item_t * )pthread_rwlock_fcfs_queue_poll(rwlock->queue);  \
    /*  \
     * As long as we did not reach the end of the queue and as long as the item   \
     * is disabled - destroy the item, dequeue it and poll the next item \
     * */   \
    while ((item != NULL) && item->is_disabled)      \
    {       \
        dequeue_and_destroy_item(); \
        item = (pthread_rwlock_fcfs_item_t * )pthread_rwlock_fcfs_queue_poll(rwlock->queue);   \
    }
    
    
/*
 * This function polls and dequeues enough items from the queue in
 * order to accept the next round of pending items. By round, I mean
 * either an arbitrary number of readers or a single writer.
 * */
static void accept_pending_items(pthread_rwlock_fcfs_t * rwlock)
{
    pthread_rwlock_fcfs_item_t * item;

    extract_first_non_disabled_item();
   
    /* The queue is empty, so we don't need to release anything */
    if (item == NULL)
    {
        return;
    }
    
    /* Check if it's a writer or a reader */
    if (item->is_writer)
    {
        /* Release the writer. */
        pthread_cond_signal( &(item->cond) );

        /* We don't need the item anymore so let's deallocate it */
        dequeue_and_destroy_item();
    }
    else
    {
        /* Release all pending readers up to the first writer or 
         * end-of-queue */
        while ((item != NULL) && (! item->is_writer))
        {
            /* Release the reader */
            pthread_cond_signal( &(item->cond) );
            
            /* We don't need the item anymore so let's deallocate it */
            dequeue_and_destroy_item();
            
            /* Move to the next pending items */
            extract_first_non_disabled_item();
        }            
    }
}

#undef dequeue_and_destroy_item
#undef extract_first_non_disabled_item

/*
 * It is possible to determine whether the thread that releases the lock is a 
 * reader or a writer according to the state of the lock. 
 * */
extern void pthread_rwlock_fcfs_release(pthread_rwlock_fcfs_t * rwlock PTHREAD_RWLOCK_FCFS_DEBUG_ARGS)
{
    pthread_mutex_lock(&(rwlock->mutex));

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
    printf("%s - Unlock!\n", id);
#endif 

    /* 
     * If there is a writer locking the lock, than obviously he is the one
     * that releases it.
     * */
    if (rwlock->is_writer)
    {
        rwlock->is_writer = 0;

        accept_pending_items(rwlock);
    }
    else
    {
        /* 
         * Croak if there are also no readers - someone is trying to release
         * a lock which he does not have...
         * */
        assert(rwlock->num_readers > 0);

        rwlock->num_readers--;

        /*
         * Check if the lock is now free of readers. If so, we can accept 
         * pending writers.
         * */

        if (rwlock->num_readers == 0)
        {
            accept_pending_items(rwlock);     
        }
    }

    pthread_mutex_unlock(&(rwlock->mutex));
}

int
    pthread_rwlock_fcfs_timed_gain_write(
        pthread_rwlock_fcfs_t * rwlock,
        const struct timespec * abstime,
        int (*continue_callback)(void * context),
        void * context
        PTHREAD_RWLOCK_FCFS_DEBUG_ARGS
        )
{
    int ret;
    
    pthread_mutex_lock(&(rwlock->mutex));

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
    printf("%s - Want Timed Lock!\n", id);
#endif

    ret = gain_write_generic(rwlock, 1, 1, abstime, continue_callback, context);

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
    if (ret == 0)
    {
        printf("%s - Lock!\n", id);
    }
    else
    {
        printf("%s - Failed Lock!\n", id);
    }
#endif

    pthread_mutex_unlock(&(rwlock->mutex));

    return ret;
}


static int gain_read_generic(
    pthread_rwlock_fcfs_t * rwlock,
    int wait_for_access,
    int is_timed,
    const struct timespec * abstime,
    int (*continue_callback)(void * context),
    void * context
    )
{
    if ((rwlock->is_writer == 0) && pthread_rwlock_fcfs_queue_is_empty(rwlock->queue))
    {
        /* Increment the number of readers. */
        rwlock->num_readers++;

        /* Now we have read access */

        return 0;
    }
    else
    {
        pthread_rwlock_fcfs_item_t * item;

        if (! wait_for_access)
        {
            return 1;
        }

        /* Let's enqueue myself */
        
        item = malloc(sizeof(pthread_rwlock_fcfs_item_t));
        /* We are a writer, so let's designate ourselves as such */
        item->is_writer = 0;

        /* We want to gain access to the lock */
        item->is_disabled = 0;
        

        /* Initialize the condition variable */
        item->cond = initial_cond_constant;

        pthread_cond_init(&(item->cond), NULL);
    
        /* Let's put ourselves in the queue, so we will be release
         * when necessary. */
        pthread_rwlock_fcfs_queue_enqueue(rwlock->queue, item);
        
        /* Wait upon the condition variable to gain access to the rwlock */
        if (! is_timed)
        {
            pthread_cond_wait(&(item->cond), &(rwlock->mutex));
        }
        else
        {
            int ret;
           
            /* 
             * We try to get a timed wait on cond. If it times-out, we ask 
             * the user if he wishes to continue.
             */
            do
            {
                ret = pthread_cond_timedwait(&(item->cond), &(rwlock->mutex), abstime);
            } while 
                (
                    (ret == ETIMEDOUT) && 
                    (continue_callback != NULL) && 
                    continue_callback(context)
                );
           
            /* We failed for some reason (probably a time out - let's disable
             * this item, so it won't be processed.
             * */
            if (ret != 0)
            {
                item->is_disabled = 1;
                return ret;
            }
        }
           
        /* Now we have a read access */
        rwlock->num_readers++;

        return 0;
    }
}

int
    pthread_rwlock_fcfs_timed_gain_read(
        pthread_rwlock_fcfs_t * rwlock,
        const struct timespec * abstime,
        int (*continue_callback)(void * context),
        void * context
        PTHREAD_RWLOCK_FCFS_DEBUG_ARGS
        )
{
    int ret;
    
    pthread_mutex_lock(&(rwlock->mutex));

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
    printf("%s - Want Timed Lock!\n", id);
#endif
    
    ret = gain_read_generic(rwlock, 1, 1, abstime, continue_callback, context);

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG    
    if (ret == 0)
    {
        printf("%s - Lock!\n", id);
    }
    else
    {
        printf("%s - Failed Lock!\n", id);
    }
#endif
    
    pthread_mutex_unlock(&(rwlock->mutex));

    return ret;
}

void pthread_rwlock_fcfs_gain_read(pthread_rwlock_fcfs_t * rwlock PTHREAD_RWLOCK_FCFS_DEBUG_ARGS)
{
    pthread_mutex_lock(&(rwlock->mutex));

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
    printf("%s - Want Lock!\n", id);
#endif

    gain_read_generic(rwlock, 1, 0, NULL, NULL, NULL);

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
    printf("%s - Lock!\n", id);
#endif

    pthread_mutex_unlock(&(rwlock->mutex));
}

int pthread_rwlock_fcfs_try_gain_read(pthread_rwlock_fcfs_t * rwlock PTHREAD_RWLOCK_FCFS_DEBUG_ARGS)
{
    int ret;
    
    pthread_mutex_lock(&(rwlock->mutex));

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
    printf("%s - Want Try Lock!\n", id);
#endif

    ret = gain_read_generic(rwlock, 0, 0, NULL, NULL, NULL);

#ifdef PTHREAD_RWLOCK_FCFS_DEBUG
    if (ret == 0)
    {
        printf("%s - Lock!\n", id);
    }
    else
    {
        printf("%s - Failed Lock!\n", id);
    }
#endif

    pthread_mutex_unlock(&(rwlock->mutex));

    return ret;
}

